#include <stdlib.h>
#include <string.h>
#include "techs.h"

static bc_t extract_bc_haplotag(FASTQRecord *rec)
{
	char *bc_str = strrchr(rec->id, ':');
	assert(bc_str != NULL);
	*bc_str = '\0';

	char *space = strchr(rec->id, ' ');  // for Long Ranger formatted FASTQs
	if (space != NULL)
		*space = '\0';

	return encode_bc(&bc_str[1], 1);
}

static bc_t extract_bc_10x(FASTQRecord *rec)
{
	char *bc_str = strrchr(rec->id, ':');
	assert(bc_str != NULL);
	*bc_str = '\0';

	char *space = strchr(rec->id, ' ');  // for Long Ranger formatted FASTQs
	if (space != NULL)
		*space = '\0';

	return encode_bc(&bc_str[1], 0);
}


static bc_t extract_bc_tellseq(FASTQRecord *rec)
{
	// Get text after whitespace	  
	char *space = strchr(rec->id, ' ');
	if (space != NULL) {
		// If startswith BX:Z: is in longranger basic format
		if (strncmp(" BX:Z:", space, 6) == 0) {
			char *bc_str = strrchr(space, ':');
			*space = '\0';
			assert(bc_str != NULL);
			*bc_str = '\0';
			return encode_bc(&bc_str[1], 0);
		}

		// If not, remove text
		*space = '\0';
	} 
	
	// Take string after final ':' as barcode
	char *bc_str = strrchr(rec->id, ':');
	assert(bc_str != NULL);
	*bc_str = '\0';
	return encode_bc(&bc_str[1], 0);
}

static bc_t extract_bc_truseq(FASTQRecord *rec)
{
	char *bc_str = rec->id[0] == '@' ? &(rec->id)[1] : rec->id;
	return atoi(bc_str);
}

static bc_t extract_bc_cptseq(FASTQRecord *rec)
{
	char *bc_str = strrchr(rec->id, ':');
	assert(bc_str != NULL);
	*bc_str = '\0';
	return atoi(&bc_str[3]);
}

static PlatformProfile profiles[] = {
	{.name = "haplotag",
	 .extract_bc = extract_bc_haplotag,
	 .bc_len = 12,
	 .many_clouds = 0,
	 .dist_thresh = 50000,
	 .error_rate = 0.001,
	 .n_density_probs = 4,
	 .density_probs = {0.6, 0.05, 0.2, 0.01}},
	
	{.name = "10x",
	 .extract_bc = extract_bc_10x,
	 .bc_len = 16,
	 .many_clouds = 0,
	 .dist_thresh = 50000,
	 .error_rate = 0.001,
	 .n_density_probs = 4,
	 .density_probs = {0.6, 0.05, 0.2, 0.01}},

	{.name = "tru",
	 .extract_bc = extract_bc_truseq,
	 .bc_len = 0,
	 .many_clouds = 1,
	 .dist_thresh = 15000,
	 .error_rate = 0.001,
	 .n_density_probs = 4,
	 .density_probs = {0.6, 0.05, 0.2, 0.01}},

	{.name = "cpt",
	 .extract_bc = extract_bc_cptseq,
	 .bc_len = 0,
	 .many_clouds = 1,
	 .dist_thresh = 3500,
	 .error_rate = 0.01,
	 .n_density_probs = 9,
	 .density_probs = {0.6, 0.01, 0.15, 0.001, 0.05, 0.001, 0.02, 0.001, 0.01}},

	{.name = "dbs",
	 .extract_bc = extract_bc_10x,
	 .bc_len = 20,
	 .many_clouds = 0,
	 .dist_thresh = 50000,
	 .error_rate = 0.001,
	 .n_density_probs = 4,
	 .density_probs = {0.6, 0.05, 0.2, 0.01}},
	
	{.name = "tellseq",
	 .extract_bc = extract_bc_tellseq,
	 .bc_len = 18,
	 .many_clouds = 0,
	 .dist_thresh = 50000,
	 .error_rate = 0.001,
	 .n_density_probs = 4,
	 .density_probs = {0.6, 0.05, 0.2, 0.01}},
	
	{.name = NULL},
};

PlatformProfile *get_platform_profile_by_name(const char *name)
{
	for (size_t i = 0; profiles[i].name != NULL; i++) {
		if (strcmp(name, profiles[i].name) == 0)
			return &profiles[i];
	}

	return NULL;
}

